# Imports
import time
import multiprocessing
import gc
import json
import datetime
import os
from importlib.metadata import version
from multiprocessing.managers import SharedMemoryManager
from pathlib import Path

# External libraries
import numpy as np
import pandas as pd
import h5py

# MEAlytics package
from .core._bandpass import *
from .core._burst_detection import *
from .core._features import *
from .core._network_burst_detection import *
from .core._plotting import *
from .core._spike_validation import *
from .core._threshold import *
from .core._utilities import *


def get_default_parameters():
    """
    Store and retrieve global parameters
    
    Returns
    -------
    parameters : dict
        Default parameters for the library.
    
    """
    parameters={
        'low cutoff' : 200,
        'high cutoff' : 3500,
        'order' : 2,
        'threshold portion' : 0.1,
        'standard deviation multiplier' : 5,
        'rms multiplier' : 5,
        'refractory period' : 0.001,
        'spike validation method' : "Noisebased",
        'exit time' : 0.001,
        'drop amplitude' : 5,
        'max drop' : 2,
        'minimal amount of spikes' : 5,
        'default interval threshold' : 100,
        'max interval threshold' : 1000,
        'burst detection kde bandwidth' : 1,
        'min channels' : 0.5,
        'thresholding method' : 'Yen',
        'nbd kde bandwidth' : 0.05,
        'remove inactive electrodes' : True,
        'activity threshold' : 0.1,
        'use multiprocessing' : False
    }

    return parameters

def _electrode_subprocess(memory_id, shape, _type, electrode, parameters):
    """
    Function that can be called to analyse a single electrode as a subprocess when using multiprocessing.

    Parameters
    ----------
    memory_id : str
        The ID of the shared memory block.
    shape : tuple
        Shape of the shared array.
    _type : type
        Type of the shared array.
    electrode : int
        The electrode to be analysed.
    parameters : dict
        Dictionary containing global paramaters. The function will extract the values needed.
    
    """

    # Load in the data from the shared memory block in the RAM
    existing_shm = multiprocessing.shared_memory.SharedMemory(name=memory_id, create=False)
    funcdata=np.ndarray(shape, _type, buffer=existing_shm.buf)

    # From all the data, select the electrode
    data=funcdata[electrode % parameters['electrode amount']]

    # Filter the data
    data=butter_bandpass_filter(data, parameters)
    
    # Calculate the threshold
    threshold_value=fast_threshold(data, parameters)
    
    # Calculate spike values
    if parameters['spike validation method']=="Noisebased":
        pass
    elif parameters['spike validation method']=='none':
        parameters['drop amplitude']=0
    else:
        raise ValueError(f"\"{parameters['spike validation method']}\" is not a valid spike validation method")
    spike_validation(data, electrode, threshold_value, parameters)
    
    # Detect the bursts
    burst_detection(data, electrode, parameters)

    data=None
    existing_shm.close()
    print(f"Processed electrode: {electrode}")

def analyse_wells(fileadress, sampling_rate, electrode_amnt, parameters={}):
    """
    Analyse an entire MEA experiment, main function of the library.

    Parameters
    ----------
    fileadress : str
        The file location of the file that is to be analyzed.
    sampling_rate : int
        The sampling rate of the MEA experiment.
    electrode_amnt : int
        The amount of electrodes per well of the MEA experiment
    parameters : dict, optional
        Parameters that can alter the analysis, if left empty, will use default parameters.

    Returns
    -------
    output : pd.dataframe
        Pandas dataframe containing features for the different wells.

    Notes
    -----
    This function calls all other functions to perform the full analysis on a MEA-dataset.
    Besides this, it also communicates the progress with the GUI.

    Always launch the function with an “if __name__ == ‘__main__’:” guard.

    Examples
    --------
    >>> from CureQ.mea import analyse_wells, get_default_parameters
    >>> 
    >>> if __name__ == '__main__':
    ...     parameters = get_default_parameters()
    ...     fileadress = 'H:/MEA_data/mea_experiment.h5'
    ...     sampling_rate = 20000
    ...     electrode_amount = 12
    ...     analyse_wells(
    ...         fileadress=fileadress,
    ...         sampling_rate=sampling_rate,
    ...         electrode_amnt=electrode_amount,
    ...         parameters=parameters
    ...     )

    """


    analysis_time=time.time()

    lib_version=version('CureQ')
    print(f"MEA Analysis Tool - Version: {lib_version}")
    print(f"Analyzing: {fileadress}")
    
    path_obj = Path(fileadress)
    parent_dir=path_obj.parent

    # Create a directory which will contain the output
    output_folder=path_obj.stem
    output_folder+="_output"
    output_folder+=f"_{datetime.date.today()}"
    output_folder+=f"_{datetime.datetime.now().strftime('%H-%M-%S')}"

    outputpath=os.path.join(parent_dir, output_folder)
    os.makedirs(outputpath)

    # Create a file to commmunicate the progress with the GUI
    progressfile=f'{os.path.split(fileadress)[0]}/progress.npy'
    np.save(progressfile, ['starting'])
    
    # Call the freeze_support function to make sure multiprocessing still works properly if the algorithm is frozen
    multiprocessing.freeze_support()

    # Open the raw data
    print("Opening the data")
    rechunk_data=False
    with h5py.File(fileadress, 'r') as h5file:
        dataset_chunks=h5file["Data/Recording_0/AnalogStream/Stream_0/ChannelData"].chunks

        datashape=h5file["Data/Recording_0/AnalogStream/Stream_0/ChannelData"].shape
        # Check if the electrode_amnt parameter is set properly
        if datashape[0]%electrode_amnt != 0:
            raise ValueError(f"The total amount of electrodes ({datashape[0]}) is not divisible by the number of electrodes per well ({electrode_amnt})")

        # Check if dataset_chunks is not None
        if dataset_chunks:
            if dataset_chunks[0] != 1:
                rechunk_data=True
            else:
                print("Data is already correctly chunked")
                rechunk_data=False
        else:
            rechunk_data=True
            
        if rechunk_data:
            print("Data is not correctly chunked yet.\nRechunking the data will allow the tool to quickly analyze large files on limited amount of RAM")
            np.save(progressfile, ['rechunking'])
            fileadress=rechunk_dataset(fileadress=fileadress, compression_method='lzf')

    # Create figures folder for output
    os.makedirs(f"{outputpath}/figures")
    
    # Create hdf5 file for output
    output_hdf_file=f"{outputpath}/output_values.h5"
    with h5py.File(output_hdf_file, 'w') as f:
        f.create_group("spike_values")
        f.create_group("burst_values")
        f.create_group("network_values")

    # Load default parameters if none were given
    if parameters=={}:
        parameters=get_default_parameters()
        
    # Save the parameters that have been given in a JSON file 
    new_values={'output path' : outputpath,
                'output hdf file' : output_hdf_file,
                'file adress' : fileadress,
                'sampling rate' : sampling_rate,
                'electrode amount' : electrode_amnt,
                'measurements' : datashape[1],
                'library version' : lib_version}
    parameters.update(new_values)

    with open(f"{outputpath}/parameters.json", 'w') as outfile:
        json.dump(parameters, outfile, indent=4)

    # Create a list of all wells
    wells=list(range(1,int(datashape[0]/electrode_amnt)+1))

    # Flag for the first iteration
    first_iteration=True

    # With multiprocessing
    if parameters['use multiprocessing']:
        # Save the data in shared memory
        print("Loading data into shared memory")
        # Create space on the RAM
        with h5py.File(fileadress, 'r') as hdf_file:
            # Access the dataset
            dataset = hdf_file["Data"]["Recording_0"]["AnalogStream"]["Stream_0"]["ChannelData"]
            np_size=dataset[:electrode_amnt].nbytes
            shape=dataset[:electrode_amnt].shape
            _type=dataset.dtype
            measurements=dataset.shape[1]
        sharedmemory=multiprocessing.shared_memory.SharedMemory(create=True, size=np_size)
        # Communicate file size with GUI
        np.save(progressfile, [(0)*electrode_amnt, datashape[0]])
        # Create a np array in the shared memory
        data_shared=np.ndarray(shape, dtype=_type, buffer=sharedmemory.buf)
        # Get the memory ID
        memory_id=sharedmemory.name
        # Clear up memory
        data=None

        # Start up a process for every single electrode
        print("Initializing processes")
        with multiprocessing.Pool(processes=electrode_amnt) as pool:
            # Iterate over all wells
            for well in wells:
                start=time.time()
                # Calculate which electrodes belong to this well
                electrodes=np.arange((well-1)*electrode_amnt, well*electrode_amnt)
                print(f"Analyzing well: {well}, consisting of electrodes: {electrodes}")

                readtime=time.time()
                # Read in the data of the well and put it into the shared memory block
                with h5py.File(fileadress, 'r') as hdf_file:
                    dataset = hdf_file["Data"]["Recording_0"]["AnalogStream"]["Stream_0"]["ChannelData"]   
                    data_shared[:]=dataset[electrodes]
                print(f"Readtime: {time.time()-readtime}")

                # Divide the tasks to the processes
                args=[(memory_id, shape, _type, electrode, parameters) for electrode in electrodes]
                pool.starmap(_electrode_subprocess, args)

                # Calculate the network bursts
                network_burst_detection([well], parameters)
                print(f"Calculated network bursts well: {well}")

                # Calculate electrode and well features
                electrode_features_df=electrode_features(well, parameters)
                well_features_df=well_features(well, parameters)
                print(f"Calculated features well: {well}")

                # If its the first iteration, create the dataframe
                if first_iteration:
                    first_iteration=False
                    electrode_features_output=copy.deepcopy(electrode_features_df)
                    output=feature_output(electrode_features_df, well_features_df)
                # If its not the first iteration, keep appending to the dataframe
                else:
                    electrode_features_output=pd.concat([electrode_features_output, copy.deepcopy(electrode_features_df)], ignore_index=False)
                    output=pd.concat([output, feature_output(electrode_features_df, well_features_df)], axis=0, ignore_index=False)
                end=time.time()
                print(f"It took {end-start} seconds to analyse well: {well}")

                # Check if the user wants to exit the analysis
                progressdata=np.load(progressfile)
                if progressdata[0]=="abort":
                    # Clean up the shared memory
                    sharedmemory.close()
                    sharedmemory.unlink()
                    pool.terminate()
                    pool.join()
                    data=None
                    del data
                    gc.collect()
                    np.save(progressfile, ["stopped"])
                    print("stopped analysis")
                    return
                
                # Communicate progression with GUI
                np.save(progressfile, [(well)*electrode_amnt, datashape[0]])

        # Clean up the shared memory
        sharedmemory.close()
        sharedmemory.unlink()

    # Without multiprocessing
    else:  
        for well in wells:
            start=time.time()
            # Calculate which electrodes belong to this well
            electrodes=np.arange((well-1)*electrode_amnt, well*electrode_amnt)
            print(f"Analyzing well: {well}, consisting of electrodes: {electrodes}")

            # Loop through all the electrodes
            for electrode in electrodes:
                with h5py.File(fileadress, 'r') as hdf_file:
                    dataset = hdf_file["Data"]["Recording_0"]["AnalogStream"]["Stream_0"]["ChannelData"]
                    # Read in the data
                    data=dataset[electrode]
                # Filter the data
                data=butter_bandpass_filter(data, parameters)
                
                # Calculate the threshold
                threshold_value=fast_threshold(data, parameters)
                
                # Calculate spike values
                if parameters['spike validation method']=="Noisebased":
                    pass
                elif parameters['spike validation method']=='none':
                    parameters['drop amplitude']=0
                else:
                    raise ValueError(f"\"{parameters['spike validation method']}\" is not a valid spike validation method")
                spike_validation(data, electrode, threshold_value, parameters)

                # Detect the bursts
                burst_detection(data, electrode, parameters)
                
                # Check if the user wants to exit the analysis
                progressdata=np.load(progressfile)
                if progressdata[0]=="abort":
                    data=None
                    del data
                    gc.collect()
                    np.save(progressfile, ["stopped"])
                    print("stopped analysis")
                    return

                # Communicate with GUI
                np.save(progressfile, [electrode+1, datashape[0]])
                print(f"Processed electrode: {electrode}")
            measurements=datashape[1]

            # Detect network bursts
            network_burst_detection([well], parameters, save_figures=True)
            print(f"Calculated network bursts well: {well}")

            # Calculate electrode and well features
            electrode_features_df=electrode_features(well, parameters)
            well_features_df=well_features(well, parameters)
            print(f"Calculated features well: {well}")

            # If its the first iteration, create the dataframe
            if first_iteration:
                first_iteration=False
                electrode_features_output=copy.deepcopy(electrode_features_df)
                output=feature_output(electrode_features_df, well_features_df)
            # If its not the first iteration, keep appending to the dataframe
            else:
                electrode_features_output=pd.concat([electrode_features_output, copy.deepcopy(electrode_features_df)], ignore_index=False)
                output=pd.concat([output, feature_output(electrode_features_df, well_features_df)], axis=0, ignore_index=False)
            end=time.time()
            print(f"It took {end-start} seconds to analyse well: {well}")
        
        # Free up RAM
        data=None
        del data
        gc.collect()

    # Save the output
    output.to_csv(f"{outputpath}/{output_folder}_Features.csv", index=False)
    electrode_features_output.to_csv(f"{outputpath}/{output_folder}_Electrode_Features.csv", index=False)
    
    # Close the analysis
    np.save(progressfile, ['done'])
    print(f"It took {time.time()-analysis_time} seconds to analyse {fileadress}")
    print("Done")
    return output


import h5py
import re
import pandas as pd
import numpy as np
from matplotlib.figure import Figure
from matplotlib import patheffects
import matplotlib.cm as cm
from scipy.ndimage import gaussian_filter
import matplotlib.colors as mcolors
import seaborn as sns

def cmap_creation():
    """
    Creates a custom colour map, to be used later while generating the heatmap.

    Parameters
    ----------
    Returns
    -------
    smooth_cmap : Custom cmap
        This enables the cmap to be calles from anywhere in the code by referencing the dictionary.

    Notes
    -----
    The colourmap is based on the live axion biosystem heatmap.
    """
    colours = [
        (0.0, "black"),   
        (1/6, "blue"),   
        (2/6, "cyan"),
        (3/6, "green"),   
        (4/6, "yellow"), 
        (5/6, "red"),     
        (1.0, "white")
    ]

    # Create a smooth LinearSegmentedColormap
    smooth_cmap = mcolors.LinearSegmentedColormap.from_list("smooth_heatmap", colours, N=256)
    return smooth_cmap

def truesort(name):
    """
    Sorts through names and ensures it's in the right order (e.g. 2 before 11).

    Parameters
    ----------
    name : str
        Name of the electrode held in the hdf5 file.

    Returns
    -------
    (well_num, electrode_num) : int, int
        These are the numbers of the wells and the numbers of the electrode, sorted.   

    Notes
    -----
    The sorted numbers allow the program to call on the right wells and electrodes to sort them in a natural way (for humans, so 2 goes before 11).
    """
    match = re.search(r'well_(\d+)_electrode_(\d+)_spikes', name)
    if match:
        well_num = int(match.group(1))         # Extract well number
        electrode_num = int(match.group(2))    # Extract electrode number
        return (well_num, electrode_num)       # Sort by well first, then electrode
    return (float('inf'), float('inf'))        # Fallback in case of mismatch

def reshape_wells(data, n_Wells, n_Electrodes, Size):
    """
    Reshapes the individual wells for easier processing.

    Parameters
    ----------
    data : df
        The data that needs to be reshaped. This is 1 row of the data. 
    n_Wells : int
        The number of wells in a plate, typically 24, 48 or 96.
    n_Electrodes : int
        The number of electrodes in a well, typically 12 of 16.
    Size : int
        Size is to give the wells a slight padding to prevent the wells from being in the corners and allowing for a better looking gaussian smoothing.

    Returns
    -------
    well_grids : list
        This is a list of the wells, so every well gets an entry, which is a 6x6 shape.

    Notes
    -----
    Currently only allows for analysis of 16 electrodes. For any other number it needs to be edited.
    """
    data_array = data.to_numpy().flatten()
    well_grids = []
    for well in range(n_Wells):
        start = well * n_Electrodes
        end = start + n_Electrodes
        # Extract and reshape single well data
        well_data = data_array[start:end].reshape(4, 4)
        expanded_well = np.zeros((Size, Size))
        expanded_well[1:(Size-1), 1:(Size-1)] = well_data
        smoothed_well = gaussian_filter(expanded_well, sigma=0.1, mode='nearest')  
        well_grids.append(smoothed_well)
    return well_grids

def data_prepper(h5_data, parameters, Vars):
    """
    Prepares the data for the heatmaps.

    Parameters
    ----------
    h5_data : str
        Path to the hdf5 file.
    parameters : dict
        This is the dictionary containing the information about the measurement.
    Vars : dict
        This is a dictionary in which all relevant variables are stored.

    Returns
    -------
    precomputed_data : df
        Dataframe that contains one row per 0.1s of measurements, this entry is the number of spikes detected in a 1s window around the row.
    n_Wells : int
        The number of wells in the plate.
    n_Electrodes : int
        The number of electrodes in one well.
    v_max : list
        The higest amount of spikes detected in the 1s window of that measurement.
    Size : int
        The size of one side of the well that should be plotted (for 16 electrodes (4x4) it's 6, leading to a well of 6x6).
    sum_final_dataframe : list
        This is a list containing the total amount of spikes per electrode.
    max_df : list
        This list contains the highest number of detected spikes in a 1s window of each electrode.
    Notes
    -----
    This functions prepares all data for the three different heatmaps.
    """
    # Reads in the data and extracts the amount of spikes in a one second window for every 0.1s of measurements
    time_df = np.arange(0,int(parameters['measurements'])/int(parameters['sampling rate']),1/Vars["fps"])

    with h5py.File(h5_data, "r") as file:
        dataset = file["spike_values"]
        sorted_matches = sorted(dataset, key=truesort)
        
        # Temp dataframe to append values of every electrode to later
        df_list = []
        df_sum_list = []  
        for filename in sorted_matches:
            well_num, electrode_num = truesort(filename)
            col_name = f"W{well_num}_E{electrode_num}"

            # Read spike timestamps
            data = file["spike_values"][filename][:, 0]
            timestamps_array = np.round(data, 6)  # Round to match time_df precision
            n_spikes = len(data)
            # Temp dataframe to append the n spikes in a frame for one electrode
            frame_counts = []

            for frame_center in time_df:
                # Define the window: from (frame_center - 0.5s) to (frame_center + 0.5s)
                window_start = frame_center - 0.5
                window_end = frame_center + 0.5

                # Count the spikes within the 1s window
                count = np.sum((timestamps_array >= window_start) & (timestamps_array < window_end))

                # Append the 'count' for the current window
                frame_counts.append(count)

            # Store processed data for this electrode
            df_list.append(pd.Series(frame_counts, name=col_name))
            df_sum_list.append(pd.Series(n_spikes, name=col_name))

    # Makes df of the n spikes in a one second frame for every 0.1s of measurements
    final_dataframe = pd.concat(df_list, axis=1)
    sum_final_dataframe = pd.concat(df_sum_list, axis=1)
    v_max = []

    for frame_center in time_df:
        # Define the window: from (frame_center - 0.5s) to (frame_center + 0.5s)
        window_start = int(frame_center - 50)
        window_end = int(frame_center + 50)
        if window_start < 0:
            window_start = 0
        if window_end > len(final_dataframe):
            window_end = int(len(final_dataframe))
        
        # Finds the vmax in a 5s window around the 'frame center', so it decides the vmax based on the maximum of the frame_center
        vmax_index = final_dataframe.iloc[window_start:window_end].max().max()

        # Append the 'vmax_index' for the current window
        v_max.append(vmax_index)


    Final_match = re.findall(r'\d+', sorted_matches[len(sorted_matches)-1])
    n_Wells = int(Final_match[0])
    n_Electrodes = int(Final_match[1])
    Size = int(np.sqrt(n_Electrodes) + 2)

    max_values = final_dataframe.max(axis=0)
    max_df = pd.DataFrame(max_values).T

    # Precompute data to load in easier later
    precomputed_data = [reshape_wells(final_dataframe.iloc[frame], n_Wells, n_Electrodes, Size) for frame in range(Vars["Num frames"])]
    return precomputed_data, n_Wells, n_Electrodes, v_max, Size, sum_final_dataframe, max_df

def make_hm(Vars, Classes, Colour_classes):    
    """
    Prepares the figure and update function for the heatmap animation.
    It does NOT create the animation object itself.
    """    
    df = Vars["df"]
    vmin = 0
    vmax = (np.array(Vars["v_max"]).max()) * 0.5
    
    # Use Figure constructor directly
    fig = Figure(figsize=(12, 9), facecolor=Vars["background_color"])
    axs = fig.subplots(Vars["Rows"], Vars["Cols"])
    fig.subplots_adjust(wspace=0.009, hspace=0.009)

    axs = axs.ravel()

    heatmaps = []
    i = 0
    for ax in axs:
        ax.set_xticks([])
        ax.set_yticks([])
        ax.set_xlim(-0.5, 3.5)
        ax.set_ylim(3.5, -0.5)
        
        for spine in ax.spines.values():
            for class_name, indices in Classes.items():
                if (i+1) in indices:
                    color = Colour_classes.get(class_name, 'grey')
                    spine.set_color(mcolors.to_rgba(color, alpha=0.6))
                    spine.set_linewidth(0.5)
                    break  
        
        hm = ax.imshow(
            np.zeros((4,4)), 
            cmap=Vars["cmap"],
            interpolation='bicubic',
            vmax=vmax,
            vmin=vmin,
            extent=[-0.5, 3.5, 3.5, -0.5],
            origin='upper'
        )
        heatmaps.append(hm)

    pos = axs[-1].get_position()
    norm = mcolors.Normalize(vmin=vmin, vmax=vmax)
    sm = cm.ScalarMappable(cmap=Vars["cmap"], norm=norm)
    cbar = fig.colorbar(sm, ax=axs.tolist(), cax=fig.add_axes([pos.x1 + 0.02, pos.y0, 0.02, (pos.height * Vars["Rows"])]))
    cbar.set_label("Activity Level", fontsize=10, color="white")
    cbar.ax.yaxis.set_tick_params(color="white")
    cbar.ax.tick_params(axis='y', colors='white')

    title = fig.suptitle("", color='white', fontsize=16, weight='bold', y=0.95)
    title.set_path_effects([patheffects.withStroke(linewidth=3, foreground='black')])

    # This is the function the animation will call
    def update(frame):
        well_grids = df[frame]
        for hm, well_data in zip(heatmaps, well_grids):
            hm.set_array(well_data)

        time_seconds = frame / Vars["fps"]
        fig.suptitle(f"Spike rate (Spikes/Sec), Time: {time_seconds:.1f}s")
        # Return the artists that were changed
        return heatmaps + [title]

    # Return the figure and the update function, but NOT the animation object
    return fig, update


def make_hm_img(Vars, Classes, Colour_classes):  
    """
    Makes a heatmap image, showing the total number of spikes in an electrode.

    Parameters
    ----------
    Vars : dict
        This is a dictionary in which all relevant variables are stored.
    Classes : dict
        This is a dictionary in which all classes (names of cell cultures) are stored.
    
    Returns
    -------
    fig: matplotlib.figure.Figure
        The figure on which everything is drawn.
    Notes
    -----
    This is a heatmap showcasing only the total activity of each electrode. It is an image, not an animation.
    """    
    df = Vars["Precomputed Max"]
    n_Electrodes = Vars["n_Electrodes"]
    fig = Figure(figsize=(12, 9), facecolor=Vars["background_color"])
    axs = fig.subplots(Vars["Rows"], Vars["Cols"])
    fig.subplots_adjust(wspace=0.009, hspace=0.009)
    axs = axs.ravel()

    fig.suptitle(f"Total spikes per electrode", fontsize=14, color="white")
    i = 0
    vmax = df.max().max()

    # Make well plots
    for i, ax in enumerate(axs):
        well_values = df.iloc[0, i * n_Electrodes : (i + 1) * n_Electrodes].to_numpy().reshape(4, 4)

        sns.heatmap(well_values, ax=ax, cmap='plasma', vmin=0, vmax=vmax, cbar=False, square=True, xticklabels=False, yticklabels=False, annot=True, fmt="d", annot_kws={"size": 6})
        
        ax.set_frame_on(True)

        for spine in ax.spines.values():
            for class_name, indices in Classes.items():
                if (i+1) in indices:
                    color = Colour_classes.get(class_name, 'grey')  
                    spine.set_visible(True)
                    spine.set_color(mcolors.to_rgba(color, alpha=0.8))
                    spine.set_linewidth(1.5)
                    break  

    fig.patch.set_facecolor(Vars["background_color"])
    return fig  

def make_hm_max_activity(Vars, Classes, Colour_classes):  
    """
    Makes a heatmap image, showing the maximum number of detected spikes in a 1 second timeframe of all electrodes.

    Parameters
    ----------
    Vars : dict
        This is a dictionary in which all relevant variables are stored.
    Classes : dict
        This is a dictionary in which all classes (names of cell cultures) are stored.
    
    Returns
    -------
    fig: matplotlib.figure.Figure
        The figure on which everything is drawn.
    Notes
    -----
    This is a heatmap showcasing only the maximum activity of each electrode in a one second window. It is an image, not an animation.
    """
    n_Electrodes = Vars["n_Electrodes"]
    max_values = Vars["max_df"].max(axis=0)
    df = pd.DataFrame(max_values).T
    
    fig = Figure(figsize=(12, 9), facecolor=Vars["background_color"])
    axs = fig.subplots(Vars["Rows"], Vars["Cols"])
    fig.subplots_adjust(wspace=0.009, hspace=0.009)
    axs = axs.ravel()

    fig.suptitle(f"Highest amount of spikes in 1 second per electrode", fontsize=14, color="white")
    i = 0
    vmax = df.max().max()

    # Make well plots
    for i, ax in enumerate(axs):
        well_values = df.iloc[0, i * n_Electrodes : (i + 1) * n_Electrodes].to_numpy().reshape(4, 4)

        sns.heatmap(well_values, ax=ax, cmap='plasma', vmin=0, vmax=vmax, cbar=False, square=True, xticklabels=False, yticklabels=False, annot=True, fmt="d", annot_kws={"size": 6})
        
        ax.set_frame_on(True)
        
        for spine in ax.spines.values():
            for class_name, indices in Classes.items():
                if (i+1) in indices:
                    color = Colour_classes.get(class_name, 'grey')  
                    spine.set_visible(True)
                    spine.set_color(mcolors.to_rgba(color, alpha=0.8))
                    spine.set_linewidth(1.5)
                    break  

    fig.patch.set_facecolor(Vars["background_color"])
    return fig  

def create_placeholder_figure(Vars):
    """
    Makes a empty heatmap image, showing only the layout of the plate.

    Parameters
    ----------
    Vars : dict
        This is a dictionary in which all relevant variables are stored.
    Returns
    -------
    fig: matplotlib.figure.Figure
        The figure on which everything is drawn.
    Notes
    -----
    """    
    fig = Figure(figsize=(12, 9), facecolor=Vars["background_color"])
    axs = fig.subplots(Vars["Rows"], Vars["Cols"])
    fig.subplots_adjust(wspace=0.009, hspace=0.009)
    axs = axs.ravel()  

    for ax in axs:
        ax.set_facecolor('black')
        ax.set_xticks([])
        ax.set_yticks([])

        ax.set_xlim(-0.5, 3.5)
        ax.set_ylim(3.5, -0.5)
        
        for spine in ax.spines.values():            
            spine.set_color(mcolors.to_rgba('grey', alpha=0.8))
            spine.set_linewidth(1)

    pos = axs[-1].get_position()
    norm = mcolors.Normalize(vmin=0, vmax=1)
    sm = cm.ScalarMappable(cmap=Vars["cmap"], norm=norm)
    cbar = fig.colorbar(sm, 
                        ax=axs.tolist(), 
                        cax=fig.add_axes([pos.x1 + 0.02, pos.y0, 0.02, (pos.height * Vars["Rows"])]))
    cbar.set_label("Activity Level", fontsize=10, color="white")
    cbar.ax.yaxis.set_tick_params(color="white")
    cbar.ax.tick_params(axis='y', colors='white')

    title = fig.suptitle("Spike rate (Spikes/Sec)", color='white', fontsize=16, weight='bold', y=0.95)
    title.set_path_effects([patheffects.withStroke(linewidth=3, foreground='black')])
    return fig


# Imports
import os
from functools import partial
import json
import webbrowser
from tkinter import *
from tkinter import filedialog
from importlib.metadata import version
import traceback
import threading

# External libraries
import h5py
import customtkinter as ctk
from CTkToolTip import *
from CTkMessagebox import CTkMessagebox
from CTkColorPicker import *

# GUI components
from ._single_electrode_view import single_electrode_view
from ._whole_well_view import whole_well_view
from ._heatmap import heatmap_frame

class select_folder_frame(ctk.CTkFrame):
    """
    Allows the user to select a dataset and outputfile to inspect the results.
    """
    def __init__(self, parent):
        super().__init__(parent)

        self.parent=parent

        self.grid_columnconfigure(0, weight=1)
        self.grid_rowconfigure(0, weight=1)

        self.folder_path=''
        self.data_path=''

        datalocation=ctk.CTkFrame(self)
        datalocation.grid(row=0, column=0, padx=10, pady=10, sticky='nsew')
        datalocation.grid_columnconfigure(1, weight=1)

        folderlabel=ctk.CTkLabel(master=datalocation, text="Folder:")
        folderlabel.grid(row=0, column=0, padx=10, pady=10, sticky='w')
        self.btn_selectfolder=ctk.CTkButton(master=datalocation, text="Select a folder", command=self.openfolder)
        self.btn_selectfolder.grid(row=0, column=1, padx=10, pady=10, sticky='nesw')

        rawfilelabel=ctk.CTkLabel(master=datalocation, text="File:")
        rawfilelabel.grid(row=1, column=0, padx=10, pady=10, sticky='w')
        self.btn_selectrawfile=ctk.CTkButton(master=datalocation, text="Select a file", command=self.openrawfile)
        self.btn_selectrawfile.grid(row=1, column=1, padx=10, pady=10, sticky='nesw')

        view_results_button=ctk.CTkButton(self, text="View results", command= lambda: self.go_to_results(parent))
        view_results_button.grid(row=2, column=0, columnspan=2, padx=10, pady=10, sticky='nesw')

        return_button=ctk.CTkButton(self, text="Return", command= lambda: parent.show_frame(self.parent.home_frame))
        return_button.grid(row=3, column=0, columnspan=2, padx=10, pady=10, sticky='nesw')

    def openfolder(self):
            resultsfolder = filedialog.askdirectory()
            if resultsfolder != '':
                self.btn_selectfolder.configure(text=resultsfolder)
                self.folder_path=resultsfolder
    
    def openrawfile(self):
            selectedfile = filedialog.askopenfilename(filetypes=[("MEA data", "*.h5")])
            if len(selectedfile)!=0:
                self.btn_selectrawfile.configure(text=os.path.split(selectedfile)[1])
                self.data_path=selectedfile

    # Check if the correct folder/file has been selected
    def go_to_results(self, parent):
        if self.folder_path == '' or self.data_path == '':
            CTkMessagebox(title="Error", message='Please select an output folder and raw datafile', icon="cancel", wraplength=400)
            return
        
        # Check for parameters json
        try:
            parameters=open(f"{self.folder_path}/parameters.json")
            parameters=json.load(parameters)
        except Exception as error:
            traceback.print_exc()
            CTkMessagebox(title='Error', message='Could not load in the results, please make sure you have selected the correct folder', icon='cancel', wraplength=400)
            return
        
        # Check the raw datafile
        try:
            with h5py.File(self.data_path, 'r') as hdf_file:
                datashape=hdf_file["Data/Recording_0/AnalogStream/Stream_0/ChannelData"].shape
        except Exception as error:
            traceback.print_exc()
            CTkMessagebox(title='Error', message='Could not load in the raw data, please make sure you have selected the correct file', icon='cancel', wraplength=400)
            return

        # If everything is good, go to the next frame
        parent.show_frame(view_results, folder=self.folder_path, rawfile=self.data_path)


class view_results(ctk.CTkFrame):
    """
    Allow the user to view the results of the MEA analysis
    """
    def __init__(self, parent, folder, rawfile):
        super().__init__(parent)

        self.folder=folder
        self.rawfile=rawfile
        self.parent=parent

        self.parent.title(f"MEAlytics - Version: {version('CureQ')} - {self.folder}")

        self.tab_frame=ctk.CTkTabview(self, anchor='nw')
        self.tab_frame.grid(column=0, row=0, sticky='nesw', pady=10, padx=10, columnspan=2)
        self.grid_columnconfigure(0, weight=1)
        self.grid_rowconfigure(0, weight=1)

        self.tab_frame.add("Single Electrode View")
        self.tab_frame.tab("Single Electrode View").grid_columnconfigure(0, weight=1)
        self.tab_frame.tab("Single Electrode View").grid_rowconfigure(0, weight=1)

        self.tab_frame.add("Whole Well View")
        self.tab_frame.tab("Whole Well View").grid_columnconfigure(0, weight=1)
        self.tab_frame.tab("Whole Well View").grid_rowconfigure(0, weight=1)

        self.tab_frame.add("Heatmap")
        self.tab_frame.tab("Heatmap").grid_columnconfigure(0, weight=1)
        self.tab_frame.tab("Heatmap").grid_rowconfigure(0, weight=1)

        # sev = single electrode view
        # wwv = whole well view

        # Load files
        self.parameters=open(f"{folder}/parameters.json")
        self.parameters=json.load(self.parameters)
        with h5py.File(rawfile, 'r') as hdf_file:
            self.datashape=hdf_file["Data/Recording_0/AnalogStream/Stream_0/ChannelData"].shape
        well_amnt = self.datashape[0]/self.parameters["electrode amount"]

        """Single electrode view"""
        self.selected_well=1
        sev_well_button_frame=ctk.CTkFrame(self.tab_frame.tab("Single Electrode View"))
        sev_well_button_frame.grid(row=0, column=0, pady=10, padx=10, sticky='nesw')

        sev_electrode_button_frame=ctk.CTkFrame(self.tab_frame.tab("Single Electrode View"))
        sev_electrode_button_frame.grid(row=0, column=1, pady=10, padx=10, sticky='nesw')
        
        # Wellbuttons
        self.xwells, self.ywells = parent.calculate_well_grid(well_amnt)
        self.sev_wellbuttons=[]
        i=1

        for y in range(self.ywells):
            for x in range(self.xwells):
                well_btn=ctk.CTkButton(master=sev_well_button_frame, text=i, command=partial(self.set_selected_well, i), height=100, width=100, font=ctk.CTkFont(size=25))
                well_btn.grid(row=y, column=x, sticky='nesw')
                self.sev_wellbuttons.append(well_btn)
                i+=1

        # Electrode buttons
        electrode_amnt = self.parameters["electrode amount"]

        electrode_layout = parent.calculate_electrode_grid(electrode_amnt)

        i = 1
        electrodebuttons=[]
        for x in range(electrode_layout.shape[0]):
            for y in range(electrode_layout.shape[1]):
                if electrode_layout[x,y]:
                    electrode_btn=ctk.CTkButton(master=sev_electrode_button_frame, text=i, command=partial(self.open_sev_tab, i), height=100, width=100, font=ctk.CTkFont(size=25))
                    electrode_btn.grid(row=x, column=y, sticky='nesw')
                    electrodebuttons.append(electrode_btn)
                    i+=1

        """Whole well view"""
        wwv_well_button_frame=ctk.CTkFrame(self.tab_frame.tab("Whole Well View"))
        wwv_well_button_frame.grid(row=0, column=0, pady=10, padx=10, sticky='nesw')

        wwv_wellbuttons=[]
        i=1

        for y in range(self.ywells):
            for x in range(self.xwells):
                well_btn=ctk.CTkButton(master=wwv_well_button_frame, text=i, command=partial(self.open_wwv_tab, i), height=100, width=100, font=ctk.CTkFont(size=25))
                well_btn.grid(row=y, column=x)
                wwv_wellbuttons.append(well_btn)
                i+=1

        """Heatmap"""
        open_hm_button = ctk.CTkButton(master=self.tab_frame.tab("Heatmap"), text='Open Heatmap', command=self.open_heatmap_thread)
        open_hm_button.grid(row=0, column=0)

        # Button to return to main menu
        return_to_main = ctk.CTkButton(master=self, text="Return to main menu", command=lambda: self.parent.show_frame(self.parent.home_frame), fg_color=parent.gray_1)
        return_to_main.grid(row=1, column=0, pady=10, padx=10, sticky='w')

        layout_warning = ctk.CTkButton(master=self, text="Warning: The well/electrode layout is auto-generated and may not match the physical plate exactly. Click here for details.", command=lambda: webbrowser.open_new("https://cureq.github.io/CureQ/supported_plates/"), fg_color=parent.gray_1)
        layout_warning.grid(row=1 , column=1, pady=10, padx=10, sticky='e')

   
    def set_selected_well(self, i):
        self.selected_well=i
        for j in range(len(self.sev_wellbuttons)):
            self.sev_wellbuttons[j].configure(fg_color=self.parent.theme["CTkButton"]["fg_color"][1])
        self.sev_wellbuttons[i-1].configure(fg_color=self.parent.theme["CTkButton"]["hover_color"][1])

    def open_sev_tab(self, electrode):
        single_electrode_view(self.parent, self.folder, self.rawfile, self.selected_well, electrode)

    def open_wwv_tab(self, well):
        whole_well_view(self.parent, self.folder, well)

    def open_heatmap_thread(self):
        hm_thread = threading.Thread(target=self.open_heatmap)
        hm_thread.start()

    def open_heatmap(self):
        heatmap_frame(master=self.tab_frame.tab("Heatmap"), parent=self.parent, datashape=self.datashape, parameters=self.parameters, xwells=self.xwells, ywells=self.ywells, folder=self.folder, tabwidget=self.tab_frame)